;; Generated package description from php-fill.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "php-fill" "1.1.1.0.20260307.232207" "Additional fill commands for PHP code editing" '((emacs "27.1")) :commit "3608ee259c5d859ffcf78a11c9586681e0731486" :authors '(("Ariel Del Valle Lozano" . "arielmazatlan@gmail.com")) :maintainer '("Ariel Del Valle Lozano" . "arielmazatlan@gmail.com") :keywords '("php" "languages" "tools" "convenience") :url "https://github.com/arielenter/php-fill.el")
